# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['supervisor',
 'supervisor.adapters',
 'supervisor.helpers',
 'supervisor.helpers.exceptions',
 'supervisor.mixins',
 'supervisor.models',
 'supervisor.reporting',
 'supervisor.workflow']

package_data = \
{'': ['*']}

install_requires = \
['chevron>=0.14.0,<0.15.0',
 'pyconfs>=0.5.5,<0.6.0',
 'pydantic>=1.8.2,<2.0.0',
 'pyyaml>=5.4.1']

entry_points = \
{'console_scripts': ['gen_models = cli.models:run']}

setup_kwargs = {
    'name': 'thoughtful',
    'version': '0.1.6',
    'description': 'Library for building Automation Digital Work Reports',
    'long_description': '# 👷 Supervisor\n\n[![pre-commit](https://github.com/thoughtful-automation/supervisor/workflows/pre-commit/badge.svg?event=push)](https://github.com/thoughtful-automation/supervisor/actions?query=workflow%3Apre-commit+event%3Apush)\n[![test](https://github.com/thoughtful-automation/supervisor/workflows/test/badge.svg?event=push)](https://github.com/thoughtful-automation/supervisor/actions?query=workflow%3Atest+event%3Apush)\n\n<img\n  title="Supervisor"\n  alt="Supervisor — Github Header"\n  width="262px"\n  height="179.5px"\n  align="right"\n  src="https://user-images.githubusercontent.com/1096881/147704110-3116d1e3-c278-45d6-b99a-209faf2b17e0.png"\n/>\n\n> 🤖 A valid digital worker must be defined and configured by a <u>manifest</u>\n> and produce a <u>work report</u> at runtime as defined by\n> TA\'s  [Department of Digital Labor](https://github.com/Thoughtful-Automation/DODL)\n> specifications and documentation.<br>\n\nTherefor, 👷 **Supervisor** is a digital worker utility that governs the\nruntime and reporting of a Digital Worker by reading it\'s manifest config,\nexecuting it\'s workflow, and producing a detailed work report at runtime.\n\n<small>Supervisor supports `Python ≥ 3.7.5`</small>\n\n## Table of Contents\n\n- [Install](#install)\n- [Documentation](#documentation)\n  - [Setup](#setup)\n  - [Rules](#rules)\n  - [Conventions](#conventions)\n  - [Advanced Usage](#advanced-usage)\n- [Contributing](#contributing)\n- [Resources](#resources)\n\n## Install\n\nInstallation options:\n\n1. From source:\n\n    ```sh\n    pip install https://github.com/Thoughtful-Automation/supervisor\n    ```\n\n> More, coming soon.\n\n## Documentation\n\nUsing `Supervisor` is easy. Below is an overview of the basic components and\nintegration points behind `Supervisor`.\n\n### Setup\n\n1. **`manifest.yaml`** &mdash; a **Manifest** is required in the root of a\n    digital worker.\n\n    The manifest file contains the configuration and workflow process\n    documentation. It is used as a "source of truth" to both dictate\n    the runtime flow of a digital worker, and report on the details of that runtime.\n\n    Each step should detail the implementation parameters and describe it\'s\n    purpose, not as an abstract concept for the workflow, but as a detailed\n    representation of the real workflow.\n\n    The manifest file should only be changed to reflect the real-world workflow\n    implementation and in tandem with the customer success team, the proper\n    business-logic for the given task.\n\n    > #### :warning: Note\n    >\n    > The `manifest.yaml` will soon be auto-generated\n    > as part of `Otto` workflow.\n    >\n    > Until then, manifests are created manually, and\n    > will be provided prior to any integration with `Supervisor`.\n\n    For a breakdown of manifest fields, see the [Manifest Schema\n    Documentation][url:manifest-schema-json]\n    in the [:books: Schema Library][url:schema-lib].\n\n    **Example Manifest:**\n\n    ```yaml\n    uid: TEST-DW1\n    # ... some more fields\n    # like name, description, etc\n    #\n    # but workflow is where you will want to focus:\n    workflow:\n      - step_id: 1\n        title: Example Macro Step 1\n        steps:\n          - step_id: 1.1\n            title: (The title will reflect the PDD)\n            description: (The description will reflect the PDD)\n            steps:\n              - step_id: 1.1.1\n                title: Another Title\n                description: Steps can have child an infinite number of child steps\n                steps: # etc\n          - step_id: 1.2\n            title: Title 1.2\n            description: |-\n              For most steps, step_id and title are the only two required params\n      # and so on... (see below for more examples)\n   ```\n\n2. **`DigitalWorker` subclass**\n\n    The digital worker must be initialized by a subclass of supervisor\'s\n    `DigitalWorker` class.\n\n    An empty `DigitalWorker` might look like this:\n\n    ``` python\n    from supervisor import DigitalWorker\n\n    class TestDW(DigitalWorker):\n        pass\n\n    # Must be instantiated as part of rcc runner\n    if __name__ == "__main__":\n      TestDW()\n    ```\n\n3. **Steps**\n\n    All `workflow` steps inside of the `manifest.yaml` must have an associated\n    step method, tagged using the `@step(1, 1)` decorator**.\n\n    Example Digital Worker Subclass with steps:\n\n    ``` python\n    from supervisor import DigitalWorker, Reporter, step\n\n    class TestDW(DigitalWorker):\n\n        @step(1):\n        def example_macro_step_1(self):\n            # Macro steps are the only non-required step methods.\n            # However, they can be used to setup class instantiation,\n            # configuration, and other common tasks for that macros set of steps.\n\n        @step(1, 1)\n        def example_step_1_1(self):\n            # Notice that step ID\'s are separated by a comma, not a period.\n            # Steps can be arbitrarily nested, so it might be @step(2, 1, 2, 3, 3)\n            # which would be equivalent to `step_id: 2.1.2.3.3` in the manifest.\n\n        @step(1, 1, 1)\n        def example_sub_step(self):\n          pass\n\n        @step(1, 2)\n        def example_step_2_same_macro(self):\n            pass\n\n        # etc.,  etc.\n\n    # Must be instantiated\n    if __name__ == "__main__":\n      TestDW()\n   ```\n\n### Rules\n\n- All steps must have a step method in the digital worker sub class\n- All steps should be tagged using the `step` decorator.\n- Implementation scripting should correlate with the step within the workflow\n  (manifest workflow). This can be delegated to a sub-class, but should be\n\n  example:\n\n  ```python\n  @step(1)\n  def get_revenue_totals(self):\n    self.stripe = Stripe()\n\n  @step(1, 1, 1)\n  def login_to_stripe(self):\n    (user, pw) = self.stripe.login()\n    return { \'user\': user }\n  ```\n\n### Conventions\n\n- Step method names are arbitrary, but by convention should be correlated to the\n  step title.\n- Order of steps is arbitrary, but by convention should be in sequential order\n  according to the manifest (or the workflow sequence)\n\n### Advanced Usage\n\nAbove covers the basic implementation details needed to get started. Below are some\nmore advanced topics that you should know while using Supervisor.\n\n#### Step Sequence\n\nThe sequence of steps are determined by the closest relative position. You can\nthink of it like an array with child steps. Supervisor\'s workflow engine starts\nat the macro-step level and begins iterating over those steps. It will execute\nthat step (`1`), then look for the next. Before it continues to a sibling step,\nit will look for children step. The next step, if it exists, will be (`1.1`), it\nwill continue down the children tree until there are no more children, and then\nit will execute all sibling steps, linearly across that depth (`1.1.1`, `1.1.2`,\n`1.1.3`). When it reaches the final child at that step, it will go back up one\nlevel (`1.2`), execute that step, and continue with the same process (of looking\nfor children, favoring stepping down over stepping across)\n\nExample step sequence:\n\n```python\n[\'1\', \'1.1\', \'1.1.1\', \'1.1.2\', \'1.2\', \'2\', \'2.1\', \'2.2\', \'3\', \'4\', \'4.1\', \'etc\']\n```\n\nYou can debug the workflow arrangement by printing the `WorkflowEngine`\ninstance. You can access the workflow instance inside of the `DigitalWorker`\nsubclass like so:\n\n```python\nprint(self.supervisor.workflow)\n# prints <DigitalWorkflow: Steps=(1, 1.1, 1.2, 1.2.1, 1.2.2, ....)>\n```\n\n#### Step Types\n\nThere are two major step types in a manifest, which are reflected in the\nway the `DigitalWorker` steps need to be implemented.\n\n1. **Sequential Steps**\n    Most steps, by default, are sequential steps. This means, when the\n    DigitalWorker is done executing that step, it simply moves on to the next\n    step.\n\n    These steps can pass data from one step to the next by returning a\n    dictionary.\n\n    For example:\n\n    ```python\n    @step(1, 1)\n    def example_step_1_1(self) -> dict:\n        # ... do some work\n        some_data = "a string"\n        return {\n          \'some_data\': some_data,\n          \'some_other_data\': 123\n        }\n\n    @step(1, 1, 1)\n    def example_step_1_1_1(self, some_data: str, some_other_data: int):\n        # because 1.1.1 is the immediate child of 1.1, this is the next\n        # sequential step. If 1.1.1 is the last child (or only child in this case),\n        # then the return value of this step would surface in step 1.2\n        print(some_data) # prints "a string"\n        print(some_other_data) # prints 123\n    ```\n\n2. **Conditional Steps**\n\n    Conditional Steps are defined much like sequential steps, except that they\n    will contain a `when_true` and/or `when_false` property. The value of these\n    properties must be the ID of some step anywhere within the workflow.\n\n    The ony implication on the implementation of this step is that it must\n    return a `True` or `False` value.\n\n    For example:\n\n    ```yaml\n    - step_id: 1.3\n      title: Does this value equal that value?\n      when_true: 1.3.2\n      when_false: 1.3.1\n    ```\n\n    ```python\n    @step(1, 3)\n    def example_step_1_3(self) -> bool:\n        # ... do some work\n        # ask the appropriate truthy question:\n        return this_value is that_value\n\n    @step(1, 3, 1)\n    def example_step_1_3_1(self): # false\n        # this will run next if this_value is not that_value\n        pass\n\n    @step(1, 3, 2)\n    def example_step_1_3_2(self): # true\n        # this will run next if this_value is that_value\n        pass\n    ```\n\n    If one of the truthy properties (`when_true` or `when_false`) are not\n    provided, Supervisor will treat it as a sequential step and move to the\n    next step in sequential order.\n\n#### Reporting\n\nThe following data is reported in each step:\n\n- **Title, Description** &mdash; from the `manifest.yaml`\n- **Timestamp** &mdash; when the step started\n- **Duration** &mdash; how long the step ran for\n- **Status** &mdash; the execution status\n- **args** &mdash; the arguments provided to a step\n- **returned** &mdash; the return value of a step\n\nYou can get a feel for what is reported by the supervisor by looking at\nSupervisors test suite [`__snapshots__/test_digitalworker.ambr`][git:test-dw-snapshot]\n\n#### Additional Reporting\n\nOften, there are other metrics, data, and details that are helpful to log within\neach step. The following utilities are provided to record adhoc data for the report:\n\n```python\ntest_data = 123\n\n# Reporter() is a singleton class that contains two\n# methods to record adhoc data with:\n\n# You can pass in that data, with no arguments,\n# for simple and intuitive results\nReporter().record(test_data)\n\n# But it\'s encouraged to provide additional context using the\n# title and description parameters:\nReporter().record(test_data,\n                title="Apples",\n                description="The number of apples eaten by Jack")\n\n# If the data is transforming from one thing to another, then\n# you can use the record_transformation method to call out the\n# transformation before and after:\nReporter().record_transformation(before="TEST", after="test")\nReporter().record_transformation(before={"A": "Dictionary"},\n                                           after={"B": "Dictionary"})\n```\n\n#### See Tests\n\nYou can see more examples and use-cases documented within the tests:\n\n- [Test Manifest][git:test-manifest]\n- [Test DigitalWorker][git:test-digitalworker]\n- [Work Report Test Snapshot][git:test-dw-snapshot]\n\n## Contributing\n\nContributions to Supervisor are welcomed!\n\nTo get started, see the [contributing guide](CONTRIBUTING.md).\n\n## Resources\n\nLinks to related code, documentation, and applications.\n\n[**:books: Schema Library**][url:schema-lib]\n\n  > The JSON-Schema-defined documents used to validate the **Manifest** and the\n  > runtime **Work Report**\n\n[**:eagle: Department of Digital\n  Labor**][url:dodl]\n\n> Documentation and Specifications for building Digital Workers in *TA\'s\n> ecosystem*, and **Empower**\n\n[**🖥 Empower**][url:dwm]\n\n> The digital Workforce Manager (*DWM*)\n\n[**:robot: Otto**][url:otto]\n\n> The build tool for Digital Workers.\n\n---\n\n<div align="center">\n\n  Made with ❤️ by\n\n  [![Thoughtful Automation](https://user-images.githubusercontent.com/1096881/141985289-317c2e72-3c2d-4e6b-800a-0def1a05f599.png)][url:ta]\n\n</div>\n\n<!--  Link References -->\n\n[url:ta]: https://www.thoughtfulautomation.com/\n[url:dwm]: https://app.thoughtfulautomation.com/\n[url:otto]: https://github.com/Thoughtful-Automation/otto\n[url:dodl]: https://github.com/Thoughtful-Automation/dodl\n[url:schema-lib]: https://github.com/Thoughtful-Automation/schemas\n[url:manifest-schema-json]:\n    https://github.com/Thoughtful-Automation/schemas/blob/main/docs/manifest.schema.md\n[git:test-dw-snapshot]:\n    https://github.com/Thoughtful-Automation/supervisor/blob/20c5cc7632c6f429206c7471fdf28fd138994901/tests/__snapshots__/test_digitalworker.ambr#L1\n[git:test-manifest]:\n    https://github.com/Thoughtful-Automation/supervisor/blob/20c5cc7632c6f429206c7471fdf28fd138994901/tests/manifests/test_digitalworke\n[git:test-digitalworker]:\n    https://github.com/Thoughtful-Automation/supervisor/blob/20c5cc7632c6f429206c7471fdf28fd138994901/tests/test_digitalworker.py\n',
    'author': 'Thoughtful Automation',
    'author_email': 'engineering@thoughtfulautomation.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://thoughtfulautomation.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
